#include <stdio.h>

// may be defined at the project level
// in the makefile
#define MOZILLA_STRICT_API

#include "nsIModule.h"
#include "nsIFactory.h"

#include "nsIComponentManager.h"
#include "nsIComponentRegistrar.h"

// use classes to handle IIDs
// classes provide methods for comparison: Equals, etc.
static const nsIID kIModuleIID = NS_IMODULE_IID;
static const nsIID kIFactoryIID = NS_IFACTORY_IID;
static const nsIID kISupportsIID = NS_ISUPPORTS_IID;
static const nsIID kIComponentRegistrarIID = NS_ICOMPONENTREGISTRAT_IID;

// generate unique ID here with uuidgen (uuid-runtime)
#define FRANCE_CID \
{ 0x894fe5e4, 0x7467, 0x42fb, \
{ 0x8c, 0x6f, 0x2a, 0x75, 0x15, 0xf9, 0x9d, 0x63}}
// 894fe5e4-7467-42fb-8c6f-2a7515f99d63

static const nsCID kFranceCID = FRANCE_CID;

class France: public nsISupports {
  private:
    nsrefcnt mRefCnt;
  public:
    France();
    virtual ~France();

    NS_IMETHOD QueryInterface(const nsIID &aIID, void **aResult);
    NS_IMETHOD_(nsrefcnt) AddRef(void);
    NS_IMETHOD_(nsrefcnt) Release(void);
};

France::France()
{
  mRefCnt = 0;
}

France::~France()
{
}

NS_IMETHODIMP
France::QueryInterface(const nsIID &aIID, 
                       void **aResult);
{
  if (aResult == NULL) {
    return NS_ERROR_NULL_POINTER;
  }
  *aResult = NULL;
  if (aIID.Equals(kISupportsIID)) {
    *aResult = (void *) this;
  }
  if (*aResult == NULL) {
    return NS_ERROR_NO_INTERFACE;
  }
  AddRef();
  return NS_OK;
}

NS_IMETHODIMP_(nsrefcnt) France::AddRef(void) 
{
  return ++mRefCnt;
}

NS_IMETHODIMP_(nsrefcnt) France::Release(void)
{
  if (--mRefCnt == 0) {
    delete this;
    return 0;
  }
  return mRefCnt;
}

// factory implementation class for component
class FranceFactory: public nsIFactory{
  private:
    nsrefcnt mRefCnt;
  public:
    FranceFactory();
    virtual ~FranceFactory();

    NS_IMETHOD QueryInterface(const nsIID &aIID, void **aResult);
    NS_IMETHOD_(nsrefcnt) AddRef(void);
    NS_IMETHOD_(nsrefcnt) Release(void);

    NS_IMETHOD CreateInstance(nsISupports *aOuter, const nsIID &iid, void * *result);
    NS_IMETHOD LockFactory(PRBool lock);
};

FranceFactory::FranceFactory()
{
  mRefCnt = 0;
}

FranceFactory::~FranceFactory()
{
}

NS_IMETHODIMP
FranceFactory::QueryInterface(const nsIID &aIID, 
                              void **aResult);
{
  if (aResult == NULL) {
    return NS_ERROR_NULL_POINTER;
  }
  *aResult = NULL;
  if (aIID.Equals(kISupportsIID)) {
    *aResult = (void *) this;
  }
  else if (aIID.Equals(kIFactoryIID)) {
    *aResult = (void *) this;
  }
  if (*aResult == NULL) {
    return NS_ERROR_NO_INTERFACE;
  }
  AddRef();
  return NS_OK;
}

NS_IMETHODIMP_(nsrefcnt) FranceFactory::AddRef(void) 
{
  return ++mRefCnt;
}

NS_IMETHODIMP_(nsrefcnt) FranceFactory::Release(void)
{
  if (--mRefCnt == 0) {
    delete this;
    return 0;
  }
  return mRefCnt;
}

NS_IMETHODIMP
FranceFactory::CreateInstance(nsISupports *aOuter,
                              const nsIID & iid,
                              void * *result)
{
  if (!result)
    return NS_ERROR_INVALID_ARG;

  France* france = new France();
  if (!france)
    return NS_ERROR_OUT_OF_MEMORY;

  nsresult rv = france->QueryInterface(iid, result);

  if (NS_FAILED(rv)) {
    *result = nsnull;
    delete sample;
  }

  return rv;
}

NS_IMETHODIMP
FranceFactory::LockFactory(PRBool lock)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

// Module implementation
class FranceModule : public nsIModule
{
  public:
    FranceModule();
    virtual ~FranceModule();

  // nsISupports method:
  NS_IMETHOD QueryInterface(const nsIID &aIID, void **aResult);
  NS_IMETHOD_(nsrefcnt) AddRef(void);
  NS_IMETHOD_(nsrefcnt) Release(void);

  // nsIModule methods:
  NS_IMETHOD GetClassObject(nsIComponentManager *aCompMgr,
                            const nsCID & aClass,
                            const nsIID & aIID,
                            void * *aResult);
  NS_IMETHOD RegisterSelf(nsIComponentManager *aCompMgr,
                          nsIFile *aLocation,
                          const char *aLoaderStr,
                          const char *aType);
  NS_IMETHOD UnregisterSelf(nsIComponentManager *aCompMgr,
                            nsIFile *aLocation,
                            const char *aLoaderStr);
  NS_IMETHOD CanUnload(nsIComponentManager *aCompMgr, 
                       PRBool *_retval);
  private:
    nsrefcnt mRefCnt;
};

FranceModule::FranceModule()
{
  mRefCnt = 0;
}

FranceModule::~FranceModule()
{
}

NS_IMETHODIMP_(nsrefcnt) FranceModule::AddRef(void) 
{
  return ++mRefCnt;
}

NS_IMETHODIMP_(nsrefcnt) FranceModule::Release(void)
{
  if (--mRefCnt == 0) {
    mRefCnt = 1; /* stabilize */
    delete this;
    return 0;
  }
  return mRefCnt;
}

NS_IMETHODIMP
FranceModule::QueryInterface(REFNSIID aIID, 
                             void** aInstancePtr);
{
  if (!aInstancePtr) {
    return NS_ERROR_NULL_POINTER;
  }

  nsISupports* foundInterface;

  if (aIID.Equals(kIModuleIID)) {
    foundInterface = (nsIModule*) this;
  } else if (aIID.Equals(kISupportsIID)) {
    foundInterface = (nsISupports*) this;
  } else {
    foundInterface = 0;
  }
  if (foundInterface) {
    foundInterface->AddRef();
    *aInstancePtr = foundInterface;
    return NS_OK;
  }

  *aInstancePtr = foundInterface();
  return NS_NOINTERFACE;
}

NS_IMETHOD 
FranceModule::GetClassObject(nsIComponentManager *aCompMgr,
                             const nsCID & aClass,
                             const nsIID & aIID,
                             void * *aResult)
{
  if (!kFranceCID.Equals(aClass)) 
    return NS_ERROR_FACTORY_NOT_REGISTERD;

  if (!result)
    return NS_ERROR_INVALID_ARG;

  FranceFactory* factory = new FranceFactory();
  if (!factory)
    return NS_ERROR_OUT_OF_MEMORY;

  nsresult rv = factory->QueryInterface(aIID, result);

  if (NS_FAILED(rv)) {
    *result = nsnull;
    delete factory;
  }
  
  return rv;
}

NS_IMETHOD 
FranceModule::RegisterSelf(nsIComponentManager *aCompMgr,
                          nsIFile *aPath,
                          const char *registryLocation,
                          const char *componentType)
{
  nsIComponentRegistrar* compReg = nsnull;
  nsresult rv = aCompMgr->QueryInterface(kIComponentRegistrarIID, (void**)&compReg);
  if (NS_FAILED(rv))
    return rv;

  rv = compReg->RegisterFactoryLocation(kFranceCID, "France Class", nsnull, aPath, registryLocation, componentType);
  compReg->release();
  return rv;
}

NS_IMETHOD 
FranceModule::UnregisterSelf(nsIComponentManager *aCompMgr,
                            nsIFile *aPath,
                            const char *registryLocation)
{
  nsIComponentRegistrar* compReg = nsnull;
  nsresult rv = aCompMgr->QueryInterface(kIComponentRegistrarIID, (void**)&compReg);
  if (NS_FAILED(rv))
    return rv;

  rv = compReg->UnregisterFactoryLocation(kFranceCID, aPath);
  compReg->Release();

  return rv;
}

NS_IMETHOD 
FranceModule::CanUnload(nsIComponentManager *aCompMgr, 
                       PRBool *okToUnload)
{
  *okToUnload = PR_FALSE; // we do not know how to unload
  return NS_OK;
}


extern "C" NS_EXPORT nsresult NSGetModule(nsIComponentManager *servMgr,
                                         nsIFile* location,
                                         nsIModule** result_cobj)
{
  nsresult rv = NS_OK;

  // Create and initialize the module instance
  FranceModule *m = new FranceModule();
  if (!m) 
    return NS_ERROR_OUT_OF_MEMORY;
  }

  // Increase refcnt and store away nsIModule interface to m in return_cobj
  rv = m->QueryInterface(kIModuleIID, (void**)return_cobj);
  if (NS_FAILED(rv)) {
    delete m;
  }
  return rv;
}

